# Copyright (c) 2021-2023 Alibaba Cloud Computing Ltd.
# SPDX-License-Identifier: MulanPSL-2.0
import os
import sys
import numpy as np
from copy import deepcopy

def fmin(
    fn,
    opt_name: str,
    knobs: list,
    baseline: dict,
    max_iteration: int,
    rule_list:list,
    sample_num: int,
    batch_size: int,
    normalize: bool
):
    """Minimize a task over a parameter space.
    More realistically: *explore* a function over a parameter space
    according to a given algorithm, allowing up to a certain number of
    function evaluations. 

    Parameters:
    ----------

    fn : callable (trial point -> loss)
        This function evaluate the recommended point and 
        use the loss result as feedback for the optimizer
    opt_name (string)   : name and data folder of this optimizer.
    max_iteration (int) : tuning max iteration
    knobs (list)        : tuning parameters.
        [
            {
                "name"      : string, parameter name, e.g. 'vm.overcommit_memory'
                "domain"    : string, parameter domain, e.g. 'sysctl'
                "range"     : list,   parameter value range, given by MIN-value and MAX-value. If field 'range' is used, definition of 'options' and 'sequence' is invalid
                "options"   : list,   parameter value range, given by Value List. If field 'options' is used, definition of 'range' and 'sequence' is invalid
                "sequence"  : list,   parameter value range, given by Sequential Value List. If field 'sequence' is used, definition of 'range' and 'options' is invalid
                "step"      : int,    parameter value adjust step. Only take effect when field 'range' is used.
                "dtype"     : string, parameter data type.
                "base"	    : string, parameter baseline config value.
            },
        ],

    baseline(dict)          : benchmark config and baseline value.
                {
                    "benchmark_field_name" : benchmark field name, such as 'Throughput' and 'Latency'
                    {
                        "base"    : list, List of baseline benchmark runing result in each time.
                        "negative"  : boolean, If negative is true, brain is supposed to reduce this benchmark field.
                        "weight"    : float, Weight if this benchmark field. 
                        "strict"    : boolean, If strict is true, worse result if this benchmark field is Unacceptable in any trade-off.
                    },
                }
    ----------
    """

    if opt_name is None:
        # bgcs is being used as the default algorithm.
        # supporting  ['bgcs', 'lamcts']
        opt_name = "bgcs"
        normalize = False

    if knobs is None:
        raise Exception(
            "There are no knobs provided."
        )

    if baseline is None:
        raise Exception(
            "There is no baseline provided."
        )

    if max_iteration is None:
        # max_iteration is set to 100 as the default.
        max_iteration = 100

    if sample_num is None:
        # sample_num is set to 20 as the default.
        sample_num = 20

    if batch_size is None:
        # batch_size is set to 5 as the default.
        batch_size = 20

    if normalize is None:
        # normalize is set to False as the default.
        normalize = False

    if rule_list is None:
        # normalize is set to False as the default.
        rule_list = dict()


    if opt_name == 'bgcs':
        from keenopt.optimization import BgcsOptim
        _ALGORITHM = BgcsOptim
    
    elif opt_name == 'lamcts':
        from keenopt.optimization import LamctsOptim
        _ALGORITHM = LamctsOptim

    elif opt_name == 'hord':
        from keenopt.optimization import HORD

        _ALGORITHM = HORD
    
    elif opt_name == 'tpe':
        from keenopt.optimization import TPE
        _ALGORITHM = TPE


    algo = _ALGORITHM(
        opt_name = opt_name, 
        knobs = knobs, 
        baseline = baseline,
        max_iteration = max_iteration,
        rule_list  = rule_list
    )


    trails = []
    for i in range(max_iteration):
        iteration, candidate, budget = algo.acquire()
        parameter_value_list = [param['value'] for param in candidate]
       
        
        score = fn(np.array(parameter_value_list))
        feedback_score = deepcopy(baseline)
        for bench_name in feedback_score.keys():
            feedback_score[bench_name] = score
        time_data_line, benchmark_value_line = algo.feedback(
            iteration = iteration, 
            bench_score = feedback_score
        )
        print(feedback_score)
        trails.append([parameter_value_list, feedback_score])
        

    return trails


    


    

    


