# Copyright (c) 2021-2023 Alibaba Cloud Computing Ltd.
# SPDX-License-Identifier: MulanPSL-2.0
import os
import sys
import random
import pickle
import re
import time
import pickle
import logging
import numpy as np
import hyperopt

from typing import List, Dict
from copy import deepcopy
from abc import ABCMeta, abstractmethod
from multiprocessing import Process, Pipe


from pySOT import strategy, surrogate
from pySOT.experimental_design import LatinHypercube
from pySOT.optimization_problems import OptimizationProblem
from poap.controller import BasicWorkerThread, ThreadController


from keenopt.sample_func.sample_func import to_unit_cube
from keenopt.sample_func.sample_func import random as randfunc
from keenopt.sample_func.sample_func import from_unit_cube
from keenopt.searchspace.searchspace import SearchSpace

from keenopt.strategy.LaMCTS import LaMCTS
from keenopt.surrogate.MCTS import MCTS

from keenopt.strategy.BFSGradientDescent import BFSGradientDescent as BGD_Strategy
from keenopt.surrogate.MLP_regression import MLPRegressSurrogate
from keenopt.searchspace.searchspace import SearchSpace


TUNE_DATA_PATH = "./"

logger = logging.getLogger("common")

def softmax(x):
    e_max = sum(np.exp(x))
    e_x = np.exp(x)
    softmax_x = e_x/e_max
    return softmax_x


def _config2pts(config: dict, knobs: list):
    pts = []
    for param in knobs:
        _value = config[param['name']]
        if param.__contains__('options'):
            pts.append(int(param['options'].index(_value)))
        else:
            pts.append(float(_value))
    return np.array([pts])

def _adjustStep(param: dict):
    """ Get step of params and adjust it.

    If step is not defined in param, use default value step = 1.

    Adjust step to limit search space smaller than Config.MAX_SEARCH_SPACE

    Args:
        param (dict): parameter dictionary
    """
    if param['dtype'] in ['string', 'str']:
        return 1

    if param['dtype'] in ['float', 'double']:
        default_step = 0.1
    else:
        default_step = 1

    try:
        step = max(default_step, param['step'])
    except Exception:
        step = default_step

    while (param['range'][1] - param['range'][0]) / step > 1000:
        step *= 2

    return step

class OptimizerUnit(metaclass=ABCMeta):
    """Tuning Optiomizer Base class

    1. Save processing data.
    2. Defination of input & output data and acquire & feedback method.
    3. Calculating loss value with weight and strict punishment.

    Attribute:
        H_config ([dict]) : 
            List of history config.  e.g.
            H_config = [
                {
                    "param1":value,
                    "param2":value
                }
            ]
        H_budget ([float]) : Budgets of each iteration.
        H_time (numpy.array) : 
            H_time count the brain running time, shape = (max_iteration, 4)
            column: time_acquire_start, time_acquire_return, time_feedback_start, time_feedback_end
        H_points (numpy.array) : Parameter value list for each iteration, shape = (max_iteration, dim)
        H_loss (numpy.array) : Final loss value for each iteration, shape = (max_iteration, )
        H_loss_parts (numpy.aray) : Parts loss value for each iteration, shape = (max_iteration, bench_size)

        sigma (float) : Weight of strict punishment.
        rho (float) : Increase rate of sigma.

    """
    def __init__(self, 
                opt_name: str, 
                max_iteration: int, 
                knobs: list, 
                baseline: dict,
                rule_list=list()):
        
        """Init optimizer instance, use tpe algorithm

        Args:
            opt_name (string)   : name and data folder of this optimizer.
            opt_type (string)   : 'tuning' or 'collect'
            max_iteration (int) : tuning max iteration
            knobs (list)        : tuning parameters.
                [
                    {
                        "name"      : string, parameter name, e.g. 'vm.overcommit_memory'
                        "domain"    : string, parameter domain, e.g. 'sysctl'
                        "range"     : list,   parameter value range, given by MIN-value and MAX-value. If field 'range' is used, definition of 'options' and 'sequence' is invalid
                        "options"   : list,   parameter value range, given by Value List. If field 'options' is used, definition of 'range' and 'sequence' is invalid
                        "sequence"  : list,   parameter value range, given by Sequential Value List. If field 'sequence' is used, definition of 'range' and 'options' is invalid
                        "step"      : int,    parameter value adjust step. Only take effect when field 'range' is used.
                        "dtype"     : string, parameter data type.
                        "base"	    : string, parameter baseline config value.
                    },
                ],
            baseline(dict)          : benchmark config and baseline value.
                {
                    "benchmark_field_name" : benchmark field name, such as 'Throughput' and 'Latency'
                    {
                        "base"    : list, List of baseline benchmark runing result in each time.
                        "negative"  : boolean, If negative is true, brain is supposed to reduce this benchmark field.
                        "weight"    : float, Weight if this benchmark field. 
                        "strict"    : boolean, If strict is true, worse result if this benchmark field is Unacceptable in any trade-off.
                    },
                }
            rules (list)            : rules for restricting the values of related parameter
                [
                    {"p1":"param_2", "p2":"param_4", "rule":">="},
                    {"p1":"param_1", "p2":"param_4", "rule":"<"}, ...
                ]
        """
        self.bench = baseline
        self.bench_size = len(self.bench.keys())
        self.fx_weight = softmax([float(baseline[bench_name]['weight']) for bench_name in baseline.keys()])

        self.knobs = knobs
        self.dim = len(self.knobs)

        self.iteration = -1
        self.max_iteration = max_iteration

        self.rules = self.verified(rule_list,knobs)

        self.H_config = []
        self.H_budget = []
        self.H_time   = np.zeros(shape=(self.max_iteration, 4), dtype=float)
        self.H_loss   = np.zeros(shape=(self.max_iteration,), dtype=float)
        self.H_points = np.zeros(shape=(0, self.dim), dtype=float)
        self.H_score  = np.zeros(shape=(self.max_iteration, self.bench_size), dtype=float)
        self.H_loss_parts = np.zeros(shape=(self.max_iteration, self.bench_size), dtype=float)

        self.opt_name = "{}[{}]".format(re.sub(r"\)", "]", re.sub(r"\(", "[", opt_name)), self.msg())
        #self.folder_path = os.path.join(Config.tuning_data_dir, self.opt_type, self.opt_name)
        self.folder_path = os.path.join(TUNE_DATA_PATH, opt_name)

        self.sigma = 1
        self.rho = 1.005 ** (500 / self.max_iteration)



    def _getLoss(self, bench_score: dict, iteration: int):
        """Calculate loss value by benchmark score

        1. Multiply with weight of bench.
        2. Invert value of bench is negative.
        3. Add punishment of "strict" is True.

        Args:
            bench_score (dict): Score dictionary of benchmark.  e.g.
                bench_score = {
                    "Throughput": [45000,45010,49002],
                    "latency99" : [99,98,100]
                }
            iteration (int) : iteration determined the strict punishment weight.

        Return:
            loss_parts (list): Score list for each bench. Final loss is the sum of loss_parts.
        """
        loss_parts = []
        for bench_index, bench_name in enumerate(self.bench.keys()):
            # TODO: save and compare each value rather than average.
            average_score = float(np.mean(bench_score[bench_name]))
            self.H_score[iteration][bench_index] = average_score
            baseline = float(np.mean(self.bench[bench_name]['base']))

            # Relative loss
            _loss = (average_score / baseline - 1) * self.fx_weight[bench_index] * 100

            # Reverse loss if bench is positive
            if not self.bench[bench_name]['negative']:
                _loss = -_loss
            
            # strict
            if self.bench[bench_name]['strict'] and _loss > 0:
                _loss = self.sigma * _loss**2 / 2

            loss_parts.append(_loss)

        self.sigma = self.sigma * self.rho
        return loss_parts

    def _rule_check(self, config):
        for rule in self.rules:
            if not eval("{} {} {}".format(config[rule[0]], rule[2], config[rule[1]])):
                return False
        return True

    def validate(self, config, budget):
        """Validate if the acquired configurations follows prior rules.
                This method will check each existing rule by _rule_check, allow a maximum of 10 retry of acquire.
        Returns:
            validation result: True if found valid config else False
            config: found config or empty dict()
        """

        # check each rule
        rule_valid = self._rule_check(config)

        if not rule_valid:
            # penalize optimizer
            # Loss is optimized to be minimized, so the penalty is to return a large positive value
            self.feedbackImpl(self.iteration, max(self.H_loss[0]*1e3, 1e3))

            # set maximum number of retry for acquiring
            rule_check_num = 10
            while rule_check_num > 0:
                config, budget = self.acquireImpl()
                rule_check_num = rule_check_num - 1
                rule_valid = self._rule_check(config)
                if rule_valid:
                    break
                else:
                    self.feedbackImpl(self.iteration, max(self.H_loss[0]*1e3, 1e3))

            if rule_check_num == 0:
                # early stop
                return False, dict(), 0

        return True, config, budget

    def verified(self,rule_list,knobs):
        """
        Args:
        rules:[["client_body_timeout@group-1", "send_timeout", ">="],...]
        knobs:[{'domain': 'nginx_conf', 'name': 'client_header_timeout@group-1', 'range': [10, 100], 'dtype': 'int', 'base': 'None'},...]
           
        """
        operators = [">",">=","==","<=","<","!="]
        param_list = [knob["name"] for knob in knobs]
        valid_rules = []
        if len(rule_list) > 0:
            for i in range(len(rule_list)):
                if rule_list[i][-1] in operators and set(rule_list[i][:2]) <= set(param_list)  :
                    valid_rules.append(rule_list[i])

        return valid_rules


    def acquire(self):
        """Acquire a candidate and budget

        This method will call acquireImple() which implementd by optimizer class.
        This method saves history data and check validation of candidate.

        Returns:
            iteration (int) : This iteration returnd.
            candidate (list): 
                Candidate config of this iteration.  e.g.
                candidate = [
                    {
                        "name"  : "Continuous_parameter",
                        "range" : [0,100],
                        "step"  : 1,
                        "dtype" : "int",
                        "value" : 50,
                    },
                    {
                        "name"   : "Discrete_parameter",
                        "options": ['0','1'],
                        "dtype"  : "int",
                        "value" : '1',
                    }
                ]
            budget (float) : budget of benchmark in this iteration.
        """
        # Before acquire
        self.iteration += 1
        if self.iteration >= self.max_iteration:
            return -1, [], 0

        self.H_time[self.iteration][0] = time.time()

        # Acquire implement
        config, budget = self.acquireImpl()

        # validate rules if exist
        if len(self.rules) > 0:
            valid, config, budget = self.validate(config, budget)
            if not valid:
                return -1, [], 0

        if len(config.keys()) == 0:
            return -1, [], 0
        # After acquire
        candidate = deepcopy(self.knobs)
        for param in candidate:
            param['value'] = config[param['name']]

        assert len(self.H_config) == self.iteration
        assert len(self.H_budget) == self.iteration

        self.H_budget.append(budget)
        self.H_config.append(config)

        pts = _config2pts(config, self.knobs)
        assert pts.shape[0] == 1 and pts.shape[1] == self.H_points.shape[1]
        self.H_points = np.concatenate((self.H_points, pts), axis=0)

        self.H_time[self.iteration][1] = time.time()

        return self.iteration, candidate, budget


    @abstractmethod
    def acquireImpl(self):
        """Sub-class is expected to implement this method to return config and budget

        Return:
            config (dict) : 
                Configuration dictionary of this iteration.  e.g.
                config = {
                    "param1":value,
                    "param2":value
                }

            budget (float) : Benchmark running budget.       
        """
        raise NotImplemented("Optimizer.acquireImpl() is not implemented.")


    def feedback(self, iteration: int, bench_score: dict):
        """Feedback a benchmark score.

        Args:
            iteration (int) : Iteration of this condidate
            bench_score (dict) : Benchmark running result score of each bench.
        """
        self.H_time[iteration][2] = time.time()
        if iteration != self.iteration:
            raise Exception("iteration mismatch, iteration wanted = {}, iteration feedback = {}".format(
                            self.iteration, iteration))

        loss_parts = self._getLoss(bench_score, iteration)
        mathematical_loss = sum(loss_parts)

        self.H_loss[iteration] = sum(loss_parts)
        self.H_loss_parts[iteration] = np.array(loss_parts)
        self.feedbackImpl(iteration, self.H_loss[iteration])

        #self.feedbackImpl(iteration, list(bench_score.values())[0][0])
        self.H_time[iteration][3] = time.time()

        if self.iteration % 10 == 1 or self.iteration == self.max_iteration - 1:
            self.__savefile()

        # get *.csv data
        time_value_list = [str(timestamp) for timestamp in self.H_time[iteration]]
        time_data_line = ",".join(time_value_list)

        benchmark_value_list = [str(np.mean(score)) for score in bench_score.values()]

        benchmark_value_list = list(filter(
            lambda x: self.bench[list(self.bench.keys())[benchmark_value_list.index(x)]]['weight'] > 0,
            benchmark_value_list
        ))
        benchmark_value_list.append(str(mathematical_loss))
        benchmark_value_line = ",".join(benchmark_value_list)

        return time_data_line, benchmark_value_line


    @abstractmethod
    def feedbackImpl(self, iteration: int, loss: float):
        """Sub-class is expected to implement this method to receive loss value

        Args:
            iteration (int) : Iteration of this iteration
            loss (float) : Final loss of this iteration
        """
        raise NotImplemented("Optimizer.feedbackImpl() is not implemented.")


    @abstractmethod
    def msg(self):
        """Get message of this optimizer.

        Returns:
            string : message of this optimizer
        """
        raise NotImplemented("Optimizer.msg() is not implemented.")


    def best(self):
        """Get bese candidate up to new

        Returns:
            candidate (dict) : candidate of this score
        """
        H_loss = self.H_loss[:len(self.H_config)]
        best_iteration = H_loss.tolist().index(min(H_loss))

        best_bench = deepcopy(self.bench)
        for i, bench_name in enumerate(self.bench.keys()):
            best_bench[bench_name]['value'] = self.H_score[best_iteration][i]

        best_config = self.H_config[best_iteration]
        best_candidate = deepcopy(self.knobs)
        for param in best_candidate:
            param['value'] = best_config[param['name']]

        return best_iteration, best_candidate, best_bench

    
    def __savefile(self):
        if not os.path.exists(self.folder_path):
            os.makedirs(self.folder_path)
        pickle.dump(self.knobs, open(os.path.join(
            self.folder_path, "knobs.pkl"), 'wb+'))
        pickle.dump(self.bench, open(os.path.join(
            self.folder_path, "bench.pkl"), 'wb+'))
        pickle.dump(self.H_time, open(os.path.join(
            self.folder_path, "time.pkl"), 'wb+'))
        pickle.dump(self.H_loss, open(os.path.join(
            self.folder_path, "loss.pkl"), 'wb+'))
        pickle.dump(self.H_score, open(os.path.join(
            self.folder_path, "score.pkl"), 'wb+'))
        pickle.dump(self.H_points, open(os.path.join(
            self.folder_path, "points.pkl"), 'wb+'))
        pickle.dump(self.H_loss_parts, open(os.path.join(
            self.folder_path, "loss_parts.pkl"), 'wb+'))
        return self.folder_path

    def assess(self, config):
        """Assess the potential performance of config, if it's against experts' knowledge, retrieve this config

        Args:
            config = {
                    "param1":value,
                    "param2":
            }
        """
        raise NotImplementedError

    def early_stop(self, N=10, threshold=0.01):
        """Stop tuning optimization if the score did not improve in successive `N` steps
        
        """
        if self.iteration < N:
            return False
        best_loss = min(self.H_loss[:self.iteration+1])
        for i in range(self.iteration - N + 1, self.iteration+1):
            if self.H_loss[i] == best_loss:
                return False
        return True

    def getDataHead(self):
        """ Get head of parameter_value.csv, score.csv and time.csv

            Return the head of *.csv data after instance already initialized, *.csv files is supposed to be saved by keentuned.
        """
        parameter_name_list = [param['name'] for param in self.knobs]
        HEAD_parameter = ",".join(parameter_name_list)

        benchmark_name_list = [bench_name for bench_name in self.bench.keys() if self.bench[bench_name]['weight'] > 0]
        benchmark_name_list.append("mathematical_loss")
        HEAD_benchmark = ",".join(benchmark_name_list)

        HEAD_time = "acquire_start,acquire_end,feedback_start,feedback_end"

        return HEAD_parameter, HEAD_benchmark, HEAD_time
    

class BOOptimizer(OptimizerUnit):
    def __init__(self, 
                 opt_name: str,
                 max_iteration: int,
                 knobs: list,
                 baseline: dict,
                 rule_list=None,
                 sample_num: int=40,
                 initialize_epoch: int=50,
                 update_epoch: int=10,
                 train_interval: int=5,
                 train_decay: int=100,
                 normalize=True,
                 sample_func=randfunc,
                 strategy=None,
                 surrogate=None):
        
        """ initialize optmizer object.

        Args:
            searchspace (keenopt.searchspace):  Parameter searchspace.
            surrogate (keenopt.surrogate):      Surrogate model.
            strategy (keenopt.strategy):        Coordinate search strategy.
            sample_func (keenopt.sample_func):  Pre-sampling function.
            sample_num (int):                   Pre-sampling iteration.
            max_iteration (int):                Tunning iteration.
            initialize_epoch(int, optional):    Epoch to initialize neural network model.
            update_epoch(int, optional):        Epoch to update neural network model. 
        """
        
        super().__init__(opt_name,  max_iteration, knobs, baseline, rule_list)
        self.init_search_space()
        
        self.strategy    = strategy
        self.surrogate   = surrogate
        
        self.sample_num    = sample_num
        self.sample_func   = sample_func
        self.update_epoch  = update_epoch
        self.train_interval = train_interval
        self.train_decay = train_decay
        self.normalize = normalize
        self.initialize_epoch = initialize_epoch
        
        self.pts_queue = None

        # assert self.pts_queue.shape[1] == self.searchspace.dim
        
        self.H_x = np.zeros(shape=(0, self.searchspace.dim))
        self.H_fx = np.zeros(shape = (0, 1))

        self.H_pred_fx = np.zeros(shape =(0, 1))
        
        self.untrain_x = np.zeros(shape = (0, self.searchspace.dim))
        self.untrain_fx = np.zeros(shape = (0, 1))
        self.model_initialized = False

    def init_search_space(self):
        """Initialize search space
        """
        parameters = {}
        for knob in self.knobs:
            parameters[knob['name']] = knob
        self.searchspace = SearchSpace(parameters)

    def acquireConfiguration(self):
        """ Get a configuration 

        If queue is empty:
            1. Training surrogate model.
            2. Generating configuration points by coordinate search strategy and save to queue 
        
        Else:
            pop a configuration points from queue and return.
        
        Returns:
            np.array: configuration point.
        """
        if self.iteration > self.max_iteration:
            return None
        
        if self.pts_queue is None:
            self.pts_queue = self.sample_func(
                pts_num = self.sample_num,
                searchspace = self.searchspace
            )
            if self.normalize:
                self.pts_queue = to_unit_cube(self.searchspace, self.pts_queue)

        # Train/Update model if (enough iterations OR enough untrained samples) AND train interval
        if ((self.iteration > self.sample_num)
            or (self.untrain_x.shape[0] >= self.sample_num)) \
                and (self.untrain_x.shape[0] >= self.train_interval):
            if not self.model_initialized:
                tic = time.time()
                self.surrogate._fit(
                    x_train = self.H_x,
                    fx_train = self.H_fx,
                    epoch = self.initialize_epoch,
                    decay=self.train_decay,
                    reset_model = True,
                    pre_train = True)
                self.model_initialized = True
            else:
                tic = time.time()
                self.surrogate._fit(
                    x_train = self.H_x,
                    fx_train = self.H_fx,
                    epoch = self.update_epoch,
                    decay = self.train_decay,
                    reset_model = False,
                    pre_train = False)
            # self.surrogate._fit(
            #     x_train = self.H_x,
            #     fx_train = self.H_fx,
            #     epoch = self.initialize_epoch,
            #     reset_model = True,
            #     pre_train = True)

            self.untrain_x = np.zeros(shape = (0, self.surrogate.x_dim))
            self.untrain_fx = np.zeros(shape = (0, self.surrogate.fx_dim))
            
        if self.pts_queue.shape[0] == 0:
            batch_points = np.zeros(shape = (0, self.H_x.shape[1]))
            
            while batch_points.shape[0] == 0:
                batch_points = self.strategy.search(
                    x = self.H_x,
                    fx = self.H_fx,
                    iteration = self.iteration,
                    surrogate = self.surrogate)
                if batch_points.shape[0] == 0:
                    tic = time.time()
                    self.surrogate._fit(
                        x_train = self.H_x,
                        fx_train = self.H_fx,
                        epoch = self.initialize_epoch,
                        reset_model = True,
                        pre_train = True)
            
            self.pts_queue = np.concatenate((self.pts_queue, batch_points), axis=0)
            #if self.normalize:
            #    self.pts_queue = to_unit_cube(self.searchspace, self.pts_queue)
        # return self.pts_queue[0,...]
        x = self.pts_queue[0]
        return x

    
    def acquireImpl(self):
        x = self.acquireConfiguration()
        if self.normalize:
            x = from_unit_cube(self.searchspace, x)
        config = self.searchspace.pts2Configuration(x)
        return config, 1.0

        
        
    def feedbackImpl(self, iteration: int, fx: float):
        """ Feedback benchmark score

        Args:
            score (float or np.array): benchmark result
        """
        if isinstance(fx, float) or isinstance(fx, int):
            fx = np.array([fx])
        if fx.shape.__len__() == 1:
            fx = np.expand_dims(fx,axis=0)
            
        # self.iteration += 1
        x = self.pts_queue[0:1,...]
        self.pts_queue = self.pts_queue[1:,...]
        
        self.H_x = np.concatenate((self.H_x, x), axis=0)
        self.H_fx = np.concatenate((self.H_fx, fx), axis=0)

        self.H_pred_fx  = np.concatenate(
            (self.H_pred_fx, self.surrogate.eval(x)), axis=0)
        self.untrain_x  = np.concatenate((self.untrain_x, x), axis=0)
        self.untrain_fx = np.concatenate((self.untrain_fx, fx), axis=0)

    
    def dump(self, dump_path: str):
        if not os.path.exists(dump_path):
            os.makedirs(dump_path)
    
        fx_path = os.path.join(dump_path,"fx.pkl")
        pickle.dump(self.H_fx, open(fx_path,'wb'))
    
        pred_fx_path = os.path.join(dump_path,"pred_fx.pkl")
        pickle.dump(self.H_pred_fx, open(pred_fx_path,'wb'))
        
        # surrogate_path = os.path.join(dump_path,"surrogate.pkl")
        # pickle.dump(self.surrogate, open(surrogate_path,'wb'))

class TPE(OptimizerUnit):
    def __init__(self, 
                 opt_name: str, 
                 max_iteration: int,
                 knobs: list, 
                 baseline: dict,
                 rule_list=None):
                         
        """Init optimizer instance, use tpe algorithm
 
        Args:
            knobs (list): tuning parameters
            max_iteration (int): tuning max iteration
        """
 
        super(TPE, self).__init__(opt_name, max_iteration, knobs, baseline, rule_list)
        
        self.trials = hyperopt.Trials()
        self.config_pipe = Pipe()
        self.loss_pipe = Pipe()
        self.process = Process(target=hyperopt.fmin, args=(
            self._observe,          # fn    : trail point -> loss
            self.__searchSpace(),   # space : search space
            hyperopt.tpe.suggest,   # algo  : `hyperopt.rand.suggest` or `hyperopt.tpe.suggest`
            self.max_iteration,     # max_evals
            None,                   # timeout: None or int
            None,                   # loss_threshold : early-stop if loss is small enough
            self.trials,            # trials : evaluation points
            np.random.RandomState(),# rstate : random seed
            False,                  # verbose
            False,                  # allow_trials_fmin
            False,                  # pass_expr_memo_ctrl
            False,                  # return_argmin: if function fmin return args min
            # points_to_evaluate: Only works if trials=None points is [{'x': 0.0, 'y': 0.0}, {'x': 1.0, 'y': 2.0}]
            None,
            1,                      # max_queue_len: speed up parallel simulatulations
            True,                   # progress bar
            None,                   # early_stop_fn: fn to early stop
        ))
        self.process.start()
 
 
    def __searchSpace(self):
        """Build hyperopt search space
        """
        search_space = {}
        for param in deepcopy(self.knobs):
            if param.__contains__('options'):
                search_space[param['name']] = hyperopt.hp.choice(
                    param['name'], param['options'])
 
            elif param.__contains__('sequence'):
                search_space[param['name']] = hyperopt.hp.choice(
                    param['name'], param['sequence'])
 
            elif param.__contains__('range') and param['dtype'] == 'int':
                step = param['step'] if param.__contains__('step') and param['step'] else 1
                while (param['range'][1] - param['range'][0]) / step >= 1000:
                    step *= 2
                chioce_table = list(
                    range(param['range'][0], param['range'][1], step))
                search_space[param['name']] = hyperopt.hp.choice(
                    param['name'], chioce_table)
 
            elif param.__contains__('range') and param['dtype'] == 'float':
                search_space[param['name']] = hyperopt.hp.uniform(
                    param['name'], param['range'][0], param['range'][1])
 
            else:
                raise Exception("unsupported parameter type!")
 
        return search_space
 
    def _observe(self, trail_point):
        """tuning target funciton
 
        Input parameters configuration called candidate, this funciton is supposed to return benchmark score.
 
        Args:
            trail_point (dict): A candidate parameter configuration
 
        Returns:
            score (float): benchmark score
        """
        for knob in deepcopy(self.knobs):
            param_name = knob['name']
            param_value = trail_point[param_name]
            if knob['dtype'] == 'int':
                trail_point[param_name] = int(param_value)
            else:
                trail_point[param_name] = param_value
        self.config_pipe[0].send(trail_point)
        loss = self.loss_pipe[1].recv()
        return loss
 
    def acquireImpl(self):
        """Acquire a candidate from this optimizer.
 
        Returns:
            int  : iteration of this condidate
            dict : candidate
        """
        return self.config_pipe[1].recv(), 1.0
 
    def feedbackImpl(self, iteration: int, loss: float):
        """Feedback a benchmark score and candidate to this optimizer
 
        Args:
            iteration (int) : Iteration of this condidate
            candidate (dict): candidate of this score
            score (float)   : benchmark running result
        """
        self.loss_pipe[0].send(loss)
 
    def msg(self):
        """Get message of this optimizer.
 
        Returns:
            string : message of this optimizer
        """
        return "TPE"
    
class LamctsOptim(BOOptimizer):
    def __init__(self,
                 opt_name: str,
                 max_iteration: int,
                 knobs: List,
                 baseline: Dict,
                 rule_list=None,
                 sample_num: int=40,
                 batch_size: int=5,
                ):
        sample_iteration = int(np.ceil(max_iteration * 0.4))
        if sample_iteration >= batch_size:
            sample_num = sample_iteration
        else:
            sample_num = batch_size
        super().__init__(opt_name, max_iteration, knobs, baseline, rule_list, sample_num)


        self.strategy = LaMCTS(
            fx_weight=np.ones(shape=(1,)),
            sample_iteration=sample_num,
            max_iteration=max_iteration,
            batch_size=batch_size,
            searchspace=self.searchspace
        )

        self.surrogate = MCTS(
            x_dim = self.searchspace.dim,
            fx_dim = 1,
        )

    def acquireImpl(self):
        return super().acquireImpl()
    
    def feedbackImpl(self, iteration: int, fx: float):
        return super().feedbackImpl(iteration, fx)

    def msg(self):
        return "Lamcts"


class BgcsOptim(BOOptimizer):
    def __init__(self,
                 opt_name: str,
                 max_iteration: int,
                 knobs: list,
                 baseline: dict,
                 rule_list=None,
                 sample_num: int = 20,
                 batch_size: int = 5,
                 normalize=False,
                 ):
        sample_iteration = int(np.ceil(max_iteration * 0.4))
        if sample_iteration >= batch_size:
            sample_num = sample_iteration
        else:
            sample_num = batch_size
        super().__init__(opt_name,  max_iteration, knobs, baseline, rule_list, sample_num,normalize=normalize)
        parameters = {}
        for knob in self.knobs:
            parameters[knob['name']] = knob
        self.searchspace = SearchSpace(parameters)

        self.strategy = BGD_Strategy(
            fx_weight=np.ones(shape=(1,)),
            sample_iteration=sample_num,
            searchspace=self.searchspace,
            max_iteration=max_iteration,
            batch_size=batch_size
        )

        self.surrogate = MLPRegressSurrogate(
            x_dim=self.searchspace.dim,
            fx_dim=1,
        )

    def acquireImpl(self):
        return super().acquireImpl()

    def feedbackImpl(self, iteration: int, fx: float):
        return super().feedbackImpl(iteration, fx)

    def msg(self):
        return "BGCS"


class Problem(OptimizationProblem):
    def __init__(self, knobs: list, max_iteration: int):
        """ Init HORD problem to solve.

        Args:
            knobs (list): parameter list.
            max_iteration (int): max tuning round.
        """
        self.dim = len(knobs)
        self.knobs = knobs
        self.max_iteration = max_iteration

        self.loss_pipe = Pipe()
        self.configuration_pipe = Pipe()

        self.buildSearchSpace(knobs)

    def buildSearchSpace(self, knobs: list):
        """ Build HORD parameter search space.

        Args:
            knobs (list): parameter list.
        """
        self.int_var = []
        self.cont_var = []
        self.search_space = {}

        self.lb = np.zeros(self.dim, dtype=np.int64)
        self.ub = np.zeros(self.dim, dtype=np.int64)

        for index, param in enumerate(knobs):
            if param.__contains__('range'):
                step = _adjustStep(param)
                options = [param['range'][0] + i * step for i in range(
                    1 + int((param['range'][1] - param['range'][0]) / step))]

            elif param.__contains__('options'):
                options = param['options']
            
            elif param.__contains__('sequence'):
                options = param['sequence']

            else:
                raise Exception("unsupported parameter type!")

            self.search_space[param['name']] = options
            self.lb[index], self.ub[index] = 0, len(options) - 1
            self.int_var.append(index)

        self.int_var = np.array(self.int_var)
        self.cont_var = np.array(self.cont_var)

    def eval(self, config):
        """ Evaluate a config

        Send a configuration to pipe, and waiting for benchmark running result.

        Args:
            config (dict): A parameter configuration.

        Returns:
            float: loss value of this configuration.
        """
        configuration = {}

        for index, params in enumerate(self.knobs):
            configuration[params['name']
                          ] = self.search_space[params['name']][int(config[index])]

        self.configuration_pipe[0].send(configuration)
        loss = self.loss_pipe[1].recv()
        return loss


class HORD(OptimizerUnit):
    def __init__(self, 
                 opt_name: str, 
                 max_iteration: int,
                 knobs: list, 
                 baseline: dict,
                 rule_list=None):

        super(HORD, self).__init__(opt_name, max_iteration, knobs, baseline, rule_list)
        self.problem = Problem(knobs, max_iteration)

        self.strategy = self.__getStrategy()
        self.surrogate_fn = self.__getSurrogate()

        self.process = Process(target=self._fit, args=(self.problem,))
        self.process.start()

    def __getSurrogate(self):
        """ choose surrogate of pySOT

        Returns:
            pySOT.surrogate: surrogate in pySOT
        """
        # Choose surrogate of HORD
        return surrogate.RBFInterpolant

    def __getStrategy(self):
        """ choose strategy for HORD

        EI Strategy can only be used in combination with GPRegressor Surrogate.

        Returns:
            pySOT.strategy: strategy in pySOT
        """
        return strategy.DYCORSStrategy

    def _fit(self, problem):
        self.controller = ThreadController()
        ''' create and reset surrogate '''
        self.surrogate = self.surrogate_fn(
            dim=problem.dim,
            lb=problem.lb,
            ub=problem.ub,
        )

        ''' reset surrogate'''
        self.controller.strategy = self.strategy(
            max_evals=problem.max_iteration,
            opt_prob=problem,
            exp_design=LatinHypercube(
                dim=problem.dim,
                num_pts=problem.dim * 2,
                iterations=200,
            ),
            surrogate=self.surrogate
        )

        ''' add extern-dataset '''

        worker = BasicWorkerThread(self.controller, problem.eval)
        self.controller.launch_worker(worker)

        ''' Main Loop '''
        result = self.controller.run()

        return result

    def acquireImpl(self):
        configuration = self.problem.configuration_pipe[1].recv()
        return configuration, 1.0

    def feedbackImpl(self, iteration, loss):
        if iteration <= self.dim * 2:
            self.sigma = 1
        self.problem.loss_pipe[0].send(loss)

    def msg(self):
        return "HORD Optimizer(v1.0)"