# Copyright (c) 2021-2023 Alibaba Cloud Computing Ltd.
# SPDX-License-Identifier: MulanPSL-2.0
import random
import numpy as np  

from copy import deepcopy
from keenopt.searchspace.options import Options
from keenopt.searchspace.variable import Variable
from keenopt.searchspace.sequence import Sequence
from keenopt.common.log import funcLog

class SearchSpace:
    @funcLog
    def __init__(self,
                 parameters: dict,
                 ):
        """ init search space

        Args:
            parameters (dict): 
                parameter information dictionary. e.g.
                parameters = {
                    "param_name1": {
                        "dtype":"string",
                        "options":['y','n'],
                    },
                    "param_name2":{
                        "dtype":"int",
                        "range":[0,100],
                        "step":1
                    },
                }
        """
        self.param_list = []
        
        for param_name, param_info in parameters.items():
            assert param_info.__contains__('options') or \
                   param_info.__contains__('range') or \
                   param_info.__contains__('sequence')
            
            if param_info.__contains__('options'):
                param_obj = Options(
                    param_name = param_name,
                    dtype = param_info['dtype'],
                    options = param_info['options'],
                )
                self.param_list.append(param_obj)
                
            if param_info.__contains__('sequence'):
                param_obj = Sequence(
                    param_name = param_name,
                    dtype = param_info['dtype'],
                    options = param_info['sequence'],
                )
                self.param_list.append(param_obj)
                
            if param_info.__contains__('range'):
                if not param_info.__contains__('step'):
                    param_info['step'] = 1
                param_obj = Variable(
                    param_name = param_name,
                    dtype = param_info['dtype'],
                    lb = param_info['range'][0],
                    ub = param_info['range'][1],
                    step = param_info['step'] 
                )
                self.param_list.append(param_obj)
        
        self.dim = len(self.param_list)


    @funcLog
    def getRandomPts(self, pts_num: int):
        """ Generating points randomly.

        Args:
            pts_num (int): Number of points to generatiing.

        Returns:
            np.array
        """
        points = np.zeros(shape=(pts_num, 0))
        for param_obj in self.param_list:
            loc = param_obj.randomValue(pts_num)
            loc = np.expand_dims(loc,1)
            points = np.concatenate([points,loc],axis=1)
        return points


    @funcLog
    def pts2Configuration(self, point: np.array):
        """ Parse point to configuration dictionary.

        Args:
            point (np.array): configuration value vec.

        Returns:
            dict: configuration dictionary.
        """
        assert point.shape.__len__() == 1
        
        configurations = {}
        for index, param_obj in enumerate(self.param_list):
            configurations[param_obj.param_name] = param_obj.getConfig(point[index])
        return configurations
    
    
    @funcLog
    def _perturbImpl(self,
                     point: np.array,
                     probabilty: float,
                     perturbation_sigma: float):
        """ Generating a point from base_point by perturbing coordinate.

        Args:
            point (np.array):           Disturbing staring point.
            probabilty (float):         Probabilty of perturbing a coordinate
            perturbation_sigma (float): Perturbation ~ N(0,perturbation_sigma)
        """
        
        need_perturb = np.random.rand(point.shape[0]) < probabilty
        for dim_index in range(point.shape[0]):
            if need_perturb[dim_index]:
                perturbation = np.random.randn(1) * perturbation_sigma
                point[dim_index] = self.param_list[dim_index].perturb(
                    value = point[dim_index],
                    perturbation = perturbation
                )
        return point
    
    
    @funcLog
    def perturbPoints(self,
                      num_pts: int,
                      points: np.array,
                      probabilty: float,
                      perturbation_sigma: float):
        """ Generating many points from base_point by perturbing coordinate.
        
        Args:
            points (np.array):    Disturbing staring point.
            probabilty (float):         Probabilty of perturbing a coordinate
            perturbation_sigma (float): Perturbation ~ N(0,perturbation_sigma)
        """
        result_points = np.zeros(shape=(0,points.shape[1]))
        for _ in range(num_pts):
            point = self._perturbImpl(point = deepcopy(points[0]),
                                      probabilty   = probabilty,
                                      perturbation_sigma = perturbation_sigma)

            point = np.expand_dims(point,axis=0)
            result_points = np.concatenate((result_points,point),axis=0)
            
        return result_points

    
    @funcLog
    def _getPerturbDim(self, grad:np.array, probability: float):
        """ Get N = len(grad) * probability lagger grad dim.

        Args:
            grad (np.array): [description]
            probability (float): [description]
        """
        variable_dim_grad = []
        sequence_dim_grad = []
        options_dim_grad  = []
        
        for x_dim_index in range(grad.shape[0]):
            if isinstance(self.param_list[x_dim_index], Variable):
                variable_dim_grad.append((x_dim_index,abs(grad[x_dim_index])))
            
            elif isinstance(self.param_list[x_dim_index], Sequence):
                sequence_dim_grad.append((x_dim_index,abs(grad[x_dim_index])))
            
            elif isinstance(self.param_list[x_dim_index], Options):
                options_dim_grad.append((x_dim_index,abs(grad[x_dim_index])))
        
        variable_dim_grad = sorted(variable_dim_grad,key = lambda x:x[1], reverse = True)
        sequence_dim_grad = sorted(sequence_dim_grad,key = lambda x:x[1], reverse = True)
        options_dim_grad  = sorted(options_dim_grad, key = lambda x:x[1], reverse = True)
        
        variable_dim_list = [i[0] for i in variable_dim_grad]
        sequence_dim_list = [i[0] for i in sequence_dim_grad]
        options_dim_list  = [i[0] for i in options_dim_grad]
        
        return variable_dim_list, sequence_dim_list, options_dim_list

    @funcLog
    def _perturbPointImpl(self,
                          point: np.array,
                          dim_list:list,
                          grad: np.array,
                          perturbation: float,
                          probability:float):
        count = 0
        for x_dim_index in dim_list:
            param_obj = self.param_list[x_dim_index]
            value = param_obj.perturbGrad(
                    point[x_dim_index], 
                    grad[x_dim_index],
                    perturbation,
            )
            
            if point[x_dim_index] == value:
                continue
            else:
                point[x_dim_index] = value
                count += 1
                
            if count >= int(len(dim_list) * probability):
                break
        return point

    @funcLog
    def perturbPointsGrad(self,
                          points: np.array,
                          grad_matrix: np.array,
                          perturbation_list: np.array,
                          probability_list: float):
        
        """ Perturbing points by gradient

        Args:
            points (np.array):      Perturbation starting points. shape = (N, searchspace.dim)
            grad_matrix (np.array): Gradient. shape = (N, searchspace.dim, fx_dim)
            
        Returns:
            np.array
        """
        assert points.shape.__len__() == 2
        assert points.shape[0] == grad_matrix.shape[0]
        assert points.shape[1] == grad_matrix.shape[1]
        assert grad_matrix.shape.__len__() == 2
        
        N = points.shape[0]
        x_dim  = grad_matrix.shape[1]
        
        result_points = np.zeros(shape=(0,x_dim))
        
        for i in range(N):
            grad = np.squeeze(grad_matrix[i,...])
            
            variable_dim_list, sequence_dim_list, options_dim_list = self._getPerturbDim(grad, probability_list[i])
            
            new_point = deepcopy(points[i,...])
            new_point = self._perturbPointImpl(
                point = new_point,
                dim_list = variable_dim_list,
                grad = grad,
                perturbation = perturbation_list[i],
                probability = probability_list[i]
            )
            new_point = self._perturbPointImpl(
                point = new_point,
                dim_list = sequence_dim_list,
                grad = grad,
                perturbation = perturbation_list[i],
                probability = probability_list[i]
            )
            new_point = self._perturbPointImpl(
                point = new_point,
                dim_list = options_dim_list,
                grad = grad,
                perturbation = perturbation_list[i],
                probability = probability_list[i]
            )
                
            new_point = np.expand_dims(new_point, axis=0)
            result_points = np.concatenate((result_points,new_point),axis=0)
        return result_points

        
    def _distanceImpl(self, 
                      source_point: np.array, 
                      target_point: np.array):
        """ Calculating distance between two points.

        Args:
            source_point (np.array): Point to calcuate distance.
            target_point (np.array): Destination point
            
        Returns:
            float: distance if two points.
        """
        assert source_point.shape[0] == target_point.shape[0]
        
        distance = 0
        for dim in range(source_point.shape[0]):
            distance += self.param_list[dim].distance(
                        source_point[dim],target_point[dim])

        return distance
    
    
    @funcLog
    def distance(self, 
                 source_points:np.array, 
                 target_points:np.array, 
                 concat = 'min'):
        """ Calculating points distance from source_points to target_points.

        Args:
            source_points (np.array): Points to calcuate distance.
            target_points (np.array): Destination points.
            concat(str):              Concatenate type. min, max or mean.
            
        Return: 
            np.array: shape = source_points.shape[0]
        """
        assert source_points.shape.__len__() == 2
        assert target_points.shape.__len__() == 2
        assert source_points.shape[1] == target_points.shape[1]
        
        distance_matrix = np.zeros(
            shape = (source_points.shape[0],target_points.shape[0]))
        
        for i in range(source_points.shape[0]):
            for j in range(target_points.shape[0]):
                distance_matrix[i][j] = self._distanceImpl(
                    source_point = source_points[i,...],
                    target_point = target_points[j,...])
                
        if concat == 'min':
            distance_matrix = np.min(distance_matrix, axis=1)
        if concat == "max":
            distance_matrix = np.max(distance_matrix, axis=1)
        if concat == "mean":
            distance_matrix = np.mean(distance_matrix, axis=1)
            
        return distance_matrix
