# Copyright (c) 2021-2023 Alibaba Cloud Computing Ltd.
# SPDX-License-Identifier: MulanPSL-2.0
import math  
import numpy as np  

from copy import deepcopy

from keenopt.searchspace.parameter import Parameter
from keenopt.common.log import funcLog

class Variable(Parameter):
    
    @funcLog
    def __init__(self, param_name, dtype, ub, lb, step):
        super().__init__(param_name, dtype)
        self.ub = ub
        self.lb = lb
        if step is None:
            self.step = 1
        else:
            self.step = step
        self.range_size = (self.ub - self.lb) / self.step
        
        assert self.ub > self.lb
        assert self.step > 0
    
    
    def randomValue(self, nums):
        if self.dtype == "int":
            return np.random.randint(self.lb,self.ub,nums)
        else:
            return np.random.uniform(self.lb,self.ub,nums)
    
    
    def getConfig(self, value):
        assert value >= self.lb and value <= self.ub
        return self._parseDtype(value)


    def perturb(self, value, perturbation):
        """ Increase perturbation for value.

        Args:
            value (float):         Parameter value.
            perturbation (float):  Perturbation range, 0 ~ 1

        Returns:
            float
        """
        delta = int(perturbation * (self.ub - self.lb) / self.step) * self.step
        perturb_value = value + delta
        perturb_value = max(perturb_value, self.lb)
        perturb_value = min(perturb_value, self.ub)
        return perturb_value
    
    @funcLog
    def perturbGrad(self, value, grad, perturbation):
        """ perturbing value by grad

        Args:
            value (float):  Parameter value.
            grad (float):   Gradient if parameter in this value.

        Returns:
            float
        """
        delta = max(1, int(perturbation * self.range_size / 100))
        if grad > 0:
            new_value = value - self.step * delta
            new_value = max(new_value, self.lb)
        else:
            new_value = value + self.step * delta
            new_value = min(new_value, self.ub)
        return new_value
    
    
    def distance(self, value1, value2):
        """ Calculate distance between value1 and value 2.

        Returns:
            float
        """
        return np.abs(value1 - value2) / (self.ub - self.lb)
