# Copyright (c) 2021-2023 Alibaba Cloud Computing Ltd.
# SPDX-License-Identifier: MulanPSL-2.0
import numpy as np

from copy import deepcopy
from abc import ABCMeta,abstractmethod

from keenopt.common.log import funcLog
from keenopt.searchspace.options import Options
from keenopt.searchspace.sequence import Sequence
from keenopt.searchspace.variable import Variable

class Surrogate(metaclass=ABCMeta):
    @funcLog
    def __init__(self,
                 x_dim: int,
                 fx_dim: int):
        
        self.x_dim = x_dim
        self.fx_dim = fx_dim
        # self._initModel()
        
        
    @abstractmethod
    def _initModel(self):
        """ initialize neruel network model.
        """
        ...


    @abstractmethod
    def eval(self,x: np.array):
        """ Evaluate a point, return prediction value.

        Args:
            x (np.array): A configuration to predict.
        """
        ...
                
                
    @abstractmethod
    def _fit(self, epoch: int):
        """ Training neural network.

        Training neural network to fitting data.
        
        Retain part of the data as a eval-data (e.g. 10 iteration) to stop training if over fitting.
        """ 
        ...


    @funcLog
    def addPoints(self,x,fx):
        """ Add training data.

        Args:
            x (np.array): configuration numpy matrix. 
                            shape = (n, searchspace.dim) or (searchspace.dim, )
                            
            y (np.array): score of the configuration.
                            shape = (n, fx_dim) or (fx_dim, )
            
        """
        if fx.shape.__len__() == 1:
            fx = np.expand_dims(fx, 0)
            
        if x.shape.__len__() == 1:
            x = np.expand_dims(x, 0)
        
        assert x.shape.__len__() == fx.shape.__len__()
        assert x.shape[0] == fx.shape[0]
        self.x = np.concatenate((self.x,x),axis=0)
        self.fx = np.concatenate((self.fx,fx),axis=0)
    
    
    @funcLog
    def calculateGradient(self, 
                          points: np.array,
                          searchspace,
                          fx_weight: np.array,
                          scale: np.array):
        
        """ Calculating gradient of points.
        
        Continuous:
            grad = f(x + step * scale) - f(x - step * scale) / scale * 2
        Discrete:
            grad = [f(x1) - f(x) + f(x2) - f(x) + ... + f(n) - f(x)] / (n - 1)
            
        Args:
            points (np.array): Points to calculate gradient. shape = n x searchspace.dim
            scale (int):       Interval width of secant line.
        
        Return:
            np.array: Gradient matrix, shape = n x searchspace.dim
        """
        fx_dim = self.eval(points[0,...]).shape[1]
        x_dim  = points.shape[1]
        
        grad_matrix = np.zeros(shape=(0, x_dim, fx_dim))
        
        for i, point in enumerate(points):
            grad = np.zeros(shape = (x_dim,fx_dim))
            for x_dim_index in range(x_dim):
                param_obj = searchspace.param_list[x_dim_index]
                if isinstance(param_obj, Variable) or isinstance(param_obj, Sequence):
                    pointA = deepcopy(point)
                    pointA[x_dim_index] += param_obj.step * max(1, int(scale[i] * param_obj.range_size / 100))
                    pointA[x_dim_index] = min(pointA[x_dim_index], param_obj.ub)
                    
                    pointB = deepcopy(point)
                    pointB[x_dim_index] -= param_obj.step * max(1, int(scale[i] * param_obj.range_size / 100))
                    pointB[x_dim_index] = max(pointB[x_dim_index], param_obj.lb)
                    grad[x_dim_index] = (self.eval(pointA) - self.eval(pointB)) / (scale[i] * 2)
                    
                if isinstance(param_obj, Options):
                    pointA = deepcopy(point)
                    delta_fx = 0
                    for j in range(param_obj.options.__len__()):
                        pointA[x_dim_index] = j
                        delta_fx += self.eval(pointA)
                    
                    delta_fx -= self.eval(point) * param_obj.options.__len__()
                    grad[x_dim_index] = delta_fx / (param_obj.options.__len__() - 1)
                
            grad = np.expand_dims(grad, axis=0)
            grad_matrix = np.concatenate((grad_matrix,grad), axis=0)
        
        grad_matrix = np.sum(grad_matrix * fx_weight, axis=2)
            
        return grad_matrix
